'''
CuteAuth EventScripts addon providing auth service.
'''
import es
import services
import services.auth
import playerlib

info = es.AddonInfo()
info.name = "CuteAuth"
info.version = "1.0"
info.author = "GODJonez"
info.description = (
    "Auth provider for everyone, from simple to very advanced configs."
)
info.basename = "cuteauth"
info.url = "http://addons.eventscripts.com/addons/view/cuteauth"

_AUTHINST = None

def load():
    '''
    Initialize addon, triggered by es_load cuteauth.
    '''
    global _AUTHINST
    
    _AUTHINST = CuteAuthService(es.getAddonPath(info.basename)+"/authdata.txt")
    _AUTHINST.registerCapability('admin_cuteauth', _AUTHINST.ROOT)
    services.register("auth", _AUTHINST)
    es.regcmd('cute', 'cuteauth/serv_cmd', 'Administer CuteAuth')
    es.regclientcmd('cute', 'cuteauth/client_cmd', 'Administer CuteAuth')

def unload():
    '''
    Remove auth service, triggered by es_unload cuteauth.
    '''
    services.unregister("auth")
    es.unregclientcmd('cute')

class CuteAuthService(services.auth.AuthorizationService):
    '''
    The auth service itself.
    '''

    strranks = {
        'ROOT': 0,
        'ADMIN': 1,
        'POWERUSER': 2,
        'USER': 4,
    }
    
    def __init__(self, authfile):
        '''
        Initialize class instance.
        '''
        self.fallbackdata = dict()
        self.authfile = authfile
        self.authdata = self._LoadAuthData(authfile)
        self.name = "cuteauth"
        if not self.authdata:
            self.authdata = {
                'users': dict(),
                'groups': dict(),
            }
            es.dbgmsg(0, "WARNING! CuteAuth could not load authdata.txt!")
    
    def registerCapability(self, auth_capability, auth_recommendedlevel):
        '''
        Register a capability with a recommended level.
        '''
        self.fallbackdata[auth_capability] = int(auth_recommendedlevel)
        es.dbgmsg(1, 'CuteAuth, capability "%s" registered for level %s' % (
            auth_capability, auth_recommendedlevel
        ))
        return True
    
    def isUseridAuthorized(self, auth_userid, auth_capability):
        '''
        Return boolean information about if the specified userid is allowed
        to use the specified capability.
        '''
        user = self.getOfflineIdentifier(auth_userid)
        data = self.isIdAuthorized(user, auth_capability)
        es.dbgmsg(1, 'CuteAuth, userid %s (%s) authed for %s = %s' % (
            auth_userid, user, auth_capability, str(data)))
        return data
    
    def getOfflineIdentifier(self, auth_userid):
        '''
        Return an identifier used to actually get the player permission.
        '''
        player = playerlib.getPlayer(int(auth_userid))
        steamid = player.get('steamid')
        if steamid in ('STEAM_ID_LAN', 'STEAM_ID_PENDING'):
            return player.get('address').split(':')[0]
        return steamid
    
    def isIdAuthorized(self, auth_identifier, auth_capability):
        '''
        Return boolean information on whether the player is allowed to use
        the specified capability, using the indetifier given by
        getOfflineIdentifier.
        '''

        es.dbgmsg(1, 'CuteAuth isIdAuthorized(%s, %s)' % (
            auth_identifier, auth_capability,
        ))
        
        # Check if we actually have any real identifier or capability
        if (not auth_identifier) or (not auth_capability):
            es.dbgmsg(1, 'CuteAuth: either parameter was empty')
        # Check if the auth_capability is known
        elif auth_capability not in self.fallbackdata:
            es.dbgmsg(1, 'CuteAuth: unknown capability')
            if auth_identifier in self.authdata['users']:
                level = self.authdata['users'][auth_identifier]['level']
                if level <= self.ADMIN:
                    return True
        # Check if the user has permission to use the capability
        elif auth_identifier in self.authdata['users']:
            es.dbgmsg(1, 'CuteAuth: user found')
            user = self.authdata['users'][auth_identifier]
            if auth_capability in user:
                # The capability has been explicitly set for this user
                es.dbgmsg(1, 'CuteAuth: explicitly set for user')
                return bool(user[auth_capability])
            # Go through all groups the user belongs to
            if user['level'] <= self.fallbackdata[auth_capability]:
                es.dbgmsg(1, 'CuteAuth: user is on higher level than'
                    ' capability')
                allowed = True
            else:
                allowed = False
            level = self.isCapabilityInGroups(
                user['groups'] | set(['_LEVEL%d' % user['level']]),
                auth_capability
            )
            es.dbgmsg(1, 'CuteAuth: group check result: %d' % level)
            if level == 2:
                # group override
                allowed = True
            return allowed and bool(level)
        else:
            # User has not been found in database, check the required auth level
            es.dbgmsg(1, 'CuteAuth: user not found')
            level = self.isCapabilityInGroups(
                ['_LEVEL%d' % self.IDENTIFIED],
                auth_capability
            )
            if level == 0:
                # explicitly denied
                es.dbgmsg(1, 'CuteAuth: denied by _LEVEL4 group')
                return False
            elif level == 2:
                # explicitly allowed
                es.dbgmsg(1, 'CuteAuth: allowed by _LEVEL4 group')
                return True
            if self.fallbackdata[auth_capability] >= self.IDENTIFIED:
                es.dbgmsg(1, 'CuteAuth: capability is marked for normal users')
                return True
        
        # If everything else fails, assume the player is not allowed
        es.dbgmsg(1, 'CuteAuth: failover disallowing')
        return False

    def isCapabilityInGroups(self, groups, auth_capability):
        '''
        Check if the capability can be find in recursive group structure.
        '''
        allowed = 1
        for group in groups:
            if group in self.authdata['groups']:
                if auth_capability in self.authdata['groups'][group]:
                    es.dbgmsg(1, 'CuteAuth: Set on group')
                    return int(
                        self.authdata['groups'][group][auth_capability]
                    ) * 2
                allowed = self.isCapabilityInGroups(
                    self.authdata['groups'][group]['groups'], auth_capability
                )
                if not allowed:
                    return 0
        return allowed
    
    def _LoadAuthData(self, sourcefile):
        '''
        Load the authentication data from file.
        '''
        authdata = None
        try:
            authfile = open(sourcefile, 'r')
            authdata = {
                'users': dict(),
                'groups': dict(),
            }
            inedit = None
            linenumber = 0
            es.dbgmsg(1, 'CuteAuth: begin reading authdata')
            for line in authfile:
                linenumber += 1
                text = line.strip()
                es.dbgmsg(1, 'CuteAuth: line %d: "%s"' % (linenumber, text))
                commands = text.split('=')
                command = commands[0].upper().strip()
                if command == 'GROUP':
                    try:
                        groupname = commands[1].strip()
                        inedit = authdata['groups'][groupname] = {
                            'groups': set(),
                        }
                        es.dbgmsg(1, 'CuteAuth, line %d, GROUP %s' % (
                            linenumber, groupname
                        ))
                    except IndexError:
                        es.dbgmsg(0, 'CuteAuth parsing error on line %d, '
                            '"=" expected' % linenumber)
                elif command == 'INGROUP':
                    try:
                        groupname = commands[1].strip()
                        inedit['groups'].add(groupname)
                        es.dbgmsg(1,
                            'CuteAuth, line %d, assigned to group %s' % (
                                linenumber, groupname
                            )
                        )
                    except IndexError:
                        es.dbgmsg(0, 'CuteAuth parsing error on line %d, '
                            '"=" expected' % linenumber)
                elif command in self.strranks or command.isdigit():
                    try:
                        userident = commands[1].strip()
                        if command in self.strranks:
                            level = self.strranks[command]
                        else:
                            level = int(command)
                        inedit = authdata['users'][userident] = {
                            'groups': set(),
                            'level': level,
                        }
                        es.dbgmsg(1, 'CuteAuth, line %d, USER %s, level %d' % (
                            linenumber, userident, level
                        ))
                    except IndexError:
                        es.dbgmsg(0, 'CuteAuth parsing error on line %d, "=" '
                            'expected' % linenumber)
                elif command == 'DENY':
                    try:
                        authname = commands[1].strip()
                        inedit[authname] = False
                        es.dbgmsg(1,
                            'CuteAuth, line %d, DENY capability %s' % (
                                linenumber, authname
                            )
                        )
                    except IndexError:
                        es.dbgmsg(0, 'CuteAuth parsing error on line %d, '
                            '"=" expected' % linenumber)
                    except TypeError:
                        es.dbgmsg(0, 'CuteAuth logic error on line %d, '
                            'unassigned capability' % linenumber)
                elif len(text)>0:
                    try:
                        inedit[text] = True
                        es.dbgmsg(1,
                            'CuteAuth, line %d, ALLOW capability %s' % (
                                linenumber, text
                            )
                        )
                    except TypeError:
                        es.dbgmsg(0, 'CuteAuth logic error on line %d,'
                            'unassigned capability' % linenumber)
                else:
                    es.dbgmsg(1, 'CuteAuth: empty line')
            es.dbgmsg(1, 'CuteAuth: stop reading datafile')    
        except IOError, error:
            es.dbgmsg(0, "CuteAuth IO error: " + str(error))
        finally:
            es.dbgmsg(1, 'CuteAuth: finally')
            try:
                authfile.close()
            except NameError:
                pass
        return authdata

    def _save(self):
        '''
        Save the data back to file.
        '''
        try:
            authfile = open(self.authfile, 'w')
            for ukey in self.authdata['users']:
                user = self.authdata['users'][ukey]
                level = user['level']
                authfile.write('%d = %s\n' % (level, ukey))
                for gkey in user['groups']:
                    authfile.write('\tINGROUP = %s\n' % gkey)
                for ckey in user:
                    if ckey != 'level' and ckey != 'groups':
                        if user[ckey]:
                            authfile.write('\t%s\n' % ckey)
                        else:
                            authfile.write('\tDENY = %s\n' % ckey)
                authfile.write('\n')
            for ukey in self.authdata['groups']:
                user = self.authdata['groups'][ukey]
                authfile.write('GROUP = %s\n' % ukey)
                for gkey in user['groups']:
                    authfile.write('\tINGROUP = %s\n' % gkey)
                for ckey in user:
                    if ckey != 'groups':
                        if user[ckey]:
                            authfile.write('\t%s\n' % ckey)
                        else:
                            authfile.write('\tDENY = %s\n' % ckey)
                authfile.write('\n')
        except IOError, error:
            es.dbgmsg(0, "CuteAuth IO error: " + str(error))
            return False
        finally:
            try:
                authfile.close()
            except NameError:
                pass
        return True
    
    def _reload(self):
        '''
        Reload auth data from the file.
        '''
        newdata = self._LoadAuthData(self.authfile)
        if newdata:
            self.authdata = newdata
            return True
        return False
    
def serv_cmd():
    '''
    Server console command handler.
    '''
    do_cmd(0, es.dbgmsg)

def client_cmd():
    '''
    Client console command handler.
    '''
    userid = int(es.getcmduserid())
    if _AUTHINST.isUseridAuthorized(userid, 'admin_cuteauth'):
        do_cmd(userid, sendecho)
    else:
        es.tell(userid, '#green', 'You do not have permission '
            'to edit CuteAuth')

def sendecho(userid, text):
    '''
    Send message to user console.
    '''
    es.cexec(userid, 'echo %s' % text)

def join_args(args, joinable):
    '''
    Join elements in args list at joinable element.
    '''
    while joinable in args:
        index = args.index(joinable)
        startidx = max((index-1, 0))
        endidx = index+2
        args[startidx:endidx] = [''.join(args[startidx:endidx])]
        
def do_cmd(userid, echo):
    '''
    Handle console command from either client or server.
    '''
    argc = es.getargc()
    args = [es.getargv(idx) for idx in xrange(argc)]
    join_args(args, ':')
    argc = len(args)
    try:
        if argc == 1:
            echo(userid, 'CuteAuth administration command usage:')
            echo(userid, 'cute <command> <parameter>')
            echo(userid, ' adduser     deluser      addgroup     delgroup')
            echo(userid, ' adduserperm deluserperm  addgroupperm delgroupperm')
            echo(userid, ' utogroup    gtogroup     ufromgroup   gfromgroup')
            echo(userid, ' save        reload')
            return
        else:
            command = args[1].lower()
        if command == 'adduser':
            if argc < 3:
                echo(userid, 'cute adduser <userid|uniqueid|ip> [level]')
                echo(userid, ' - adds specified user to auth database')
                echo(userid, ' - level: %d=ADMIN, %d=POWERUSER, %d=USER' % (
                    _AUTHINST.ADMIN, _AUTHINST.POWERUSER, _AUTHINST.IDENTIFIED))
                echo(userid, ' - defaults to level %d (ADMIN)' % _AUTHINST.ADMIN)
            else:
                if argc >= 4:
                    level = int(args[3])
                else:
                    level = _AUTHINST.ADMIN
                uid = args[2]
                if uid.isdigit():
                    uid = _AUTHINST.getOfflineIdentifier(uid)
                if uid in _AUTHINST.authdata['users']:
                    echo(userid, 'Level changed for existing user %s' % uid)
                    _AUTHINST.authdata['users'][uid]['level'] = level
                else:
                    _AUTHINST.authdata['users'][uid] = {
                        'groups': set(),
                        'level': level,
                    }
                    echo(userid, 'User %s added' % uid)
        elif command == 'deluser':
            if argc < 3:
                echo(userid, 'cute deluser <userid|uniqueid|ip>')
                echo(userid, ' - deletes specified user from auth database')
            else:
                uid = args[2]
                if uid.isdigit():
                    uid = _AUTHINST.getOfflineIdentifier(uid)
                if uid in _AUTHINST.authdata['users']:
                    del _AUTHINST.authdata['users'][uid]
                    echo(userid, 'User %s deleted' % uid)
                else:
                    echo(userid, 'User %s not found' % uid)
        elif command == 'addgroup':
            if argc < 3:
                echo(userid, 'cute addgroup <groupname>')
                echo(userid, ' - creates a new group')
            else:
                group = args[2]
                if group in _AUTHINST.authdata['groups']:
                    echo(userid, 'Group %s already exists' % group)
                else:
                    _AUTHINST.authdata['groups'][group] = {
                        'groups': set(),
                        }
                    echo(userid, 'Group %s added' % group)
        elif command == 'delgroup':
            if argc < 3:
                echo(userid, 'cute delgroup <groupname>')
                echo(userid, ' - deletes a group')
            else:
                group = args[2]
                if group in _AUTHINST.authdata['groups']:
                    del _AUTHINST.authdata['groups'][group]
                    echo(userid, 'Group %s deleted' % group)
                else:
                    echo(userid, 'Group %s not found' % group)
        elif command == 'adduserperm':
            if argc < 4:
                echo(userid, 'cute adduserperm <userid|uniqueid|ip>'
                    ' <permission> [0]')
                echo(userid, ' - adds new permission/capability for user')
                echo(userid, ' - optional 0 parameter denies this permission')
            else:
                if argc >= 5:
                    level = bool(int(args[4]))
                else:
                    level = True
                uid = args[2]
                perm = args[3]
                if uid.isdigit():
                    uid = _AUTHINST.getOfflineIdentifier(uid)
                if uid in _AUTHINST.authdata['users']:
                    _AUTHINST.authdata['users'][uid][perm] = level
                    if level:
                        echo(userid, 'User %s is now allowed to %s' % (
                            uid, perm
                        ))
                    else:
                        echo(userid, 'User %s is now denied to %s' % (
                            uid, perm
                        ))
                else:
                    echo(userid, 'User %s not found' % uid)
        elif command == 'deluserperm':
            if argc < 4:
                echo(userid, 'cute deluserperm <userid|uniqueid|ip>'
                    ' <permission>')
                echo(userid, ' - deletes a permission/capability from user')
            else:
                uid = args[2]
                perm = args[3]
                if uid.isdigit():
                    uid = _AUTHINST.getOfflineIdentifier(uid)
                if uid in _AUTHINST.authdata['users']:
                    del _AUTHINST.authdata['users'][uid][perm]
                    echo(userid, 'Permission %s deleted from %s' % (perm, uid))
                else:
                    echo(userid, 'User %s not found' % uid)
        elif command == 'addgroupperm':
            if argc < 4:
                echo(userid, 'cute addgroupperm <groupname> <permission> [0]')
                echo(userid, ' - adds new permission/capability for user')
                echo(userid, ' - optional 0 parameter denies this permission')
            else:
                if argc >= 5:
                    level = bool(int(args[4]))
                else:
                    level = True
                group = args[2]
                perm = args[3]
                if group in _AUTHINST.authdata['groups']:
                    _AUTHINST.authdata['groups'][group][perm] = level
                    if level:
                        echo(userid, 'Group %s is now allowed to %s' % (
                            group, perm
                        ))
                    else:
                        echo(userid, 'Group %s is now denied to %s' % (
                            group, perm
                        ))
                else:
                    echo(userid, 'Group %s not found' % group)
        elif command == 'delgroupperm':
            if argc < 4:
                echo(userid, 'cute delgroupperm <groupname> <permission>')
                echo(userid, ' - deletes a permission/capability from group')
            else:
                group = args[2]
                perm = args[3]
                if group in _AUTHINST.authdata['groups']:
                    del _AUTHINST.authdata['groups'][group][perm]
                    echo(userid, 'Permission %s deleted from %s' % (
                        perm, group
                    ))
                else:
                    echo(userid, 'Group %s not found' % group)
        elif command == 'utogroup':
            if argc < 4:
                echo(userid, 'cute utogroup <userid|uniqueid|ip> <groupname>')
                echo(userid, ' - assigns user to a group')
            else:
                uid = args[2]
                group2 = args[3]
                if uid.isdigit():
                    uid = _AUTHINST.getOfflineIdentifier(uid)
                if uid in _AUTHINST.authdata['users']:
                    _AUTHINST.authdata['users'][uid]['groups'].add(group2)
                    echo(userid, 'User %s is now in %s' % (uid, group2))
                else:
                    echo(userid, 'User %s not found' % uid)
        elif command == 'gtogroup':
            if argc < 4:
                echo(userid, 'cute gtogroup <groupname> <groupname>')
                echo(userid, ' - assigns group to a group')
            else:
                group = args[2]
                group2 = args[3]
                if group in _AUTHINST.authdata['groups']:
                    _AUTHINST.authdata['groups'][group]['groups'].add(group2)
                    echo(userid, 'Group %s is now in %s' % (group, group2))
                else:
                    echo(userid, 'Group %s not found' % group)
        elif command == 'ufromgroup':
            if argc < 4:
                echo(userid, 'cute ufromgroup <userid|uniqueid|ip> <groupname>')
                echo(userid, ' - removes user from a group')
            else:
                uid = args[2]
                group2 = args[3]
                if uid.isdigit():
                    uid = _AUTHINST.getOfflineIdentifier(uid)
                if uid in _AUTHINST.authdata['users']:
                    if group2 in _AUTHINST.authdata['users'][uid]['groups']:
                        _AUTHINST.authdata['users'][uid]['groups'].remove(group2)
                        echo(userid, 'User %s is no longer in %s' % (
                            uid, group2
                        ))
                    else:
                        echo(userid, 'User %s was not in %s' % (uid, group2))
                else:
                    echo(userid, 'User %s not found' % uid)
        elif command == 'gfromgroup':
            if argc < 4:
                echo(userid, 'cute gfromgroup <groupname> <groupname>')
                echo(userid, ' - removes group from a group')
            else:
                group = args[2]
                group2 = args[3]
                if group in _AUTHINST.authdata['groups']:
                    if group2 in _AUTHINST.authdata['groups'][group]['groups']:
                        _AUTHINST.authdata['groups'][group]['groups'].remove(
                            group2
                        )
                        echo(userid, 'Group %s is no longer in %s' % (
                            group, group2
                        ))
                    else:
                        echo(userid, 'Group %s was not in %s' % (
                            group, group2
                        ))
                else:
                    echo(userid, 'Group %s not found' % group)
        elif command == 'save':
            if _AUTHINST._save():
                echo(userid, 'Data saved')
            else:
                echo(userid, 'Error saving data!')
        elif command == 'reload':
            if _AUTHINST._reload():
                echo(userid, 'Data reloaded')
            else:
                echo(userid, 'Error loading data!')
        else:
            echo(userid, 'Unknown subcommand %s' % command)
    except playerlib.UseridError:
        echo(userid, 'Invalid userid')
    
